@@ -33,6 +33,7 @@ import kombu.messaging
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import eventletutils
+from oslo_utils import importutils
 import six
 import six.moves
 from six.moves.urllib import parse
@@ -46,10 +47,21 @@ from oslo_messaging._drivers import pool
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
+eventlet = importutils.try_import('eventlet')
+if eventlet and eventletutils.is_monkey_patched("thread"):
+    # Here we initialize module with the native python threading module
+    # if it was already monkey patched by eventlet/greenlet.
+    stdlib_threading = eventlet.patcher.original('threading')
+else:
+    # Manage the case where we run this driver in a non patched environment
+    # and where user even so configure the driver to run heartbeat through
+    # a python thread, if we don't do that when the heartbeat will start
+    # we will facing an issue by trying to override the threading module.
+    stdlib_threading = threading
+
 # NOTE(sileht): don't exists in py2 socket module
 TCP_USER_TIMEOUT = 18
 
-
 rabbit_opts = [
     cfg.BoolOpt('ssl',
                 default=False,
@@ -76,6 +88,15 @@ rabbit_opts = [
                deprecated_name='kombu_ssl_ca_certs',
                help='SSL certification authority file '
                     '(valid only if SSL enabled).'),
+    cfg.BoolOpt('heartbeat_in_pthread',
+                default=False,
+                help="EXPERIMENTAL: Run the health check heartbeat thread "
+                     "through a native python thread. By default if this "
+                     "option isn't provided the  health check heartbeat will "
+                     "inherit the execution model from the parent process. By "
+                     "example if the parent process have monkey patched the "
+                     "stdlib by using eventlet/greenlet then the heartbeat "
+                     "will be run through a green thread."),
     cfg.FloatOpt('kombu_reconnect_delay',
                  default=1.0,
                  deprecated_group='DEFAULT',
@@ -147,6 +168,12 @@ rabbit_opts = [
                default=2,
                help='How often times during the heartbeat_timeout_threshold '
                'we check the heartbeat.'),
+    cfg.IntOpt('direct_mandatory_flag',
+               default=True,
+               help='Enable/Disable the RabbitMQ mandatory flag '
+               'for direct send. The direct send is used as reply, '
+               'so the MessageUndeliverable exception is raised '
+               'in case the client queue does not exist.'),
 ]
 
 LOG = logging.getLogger(__name__)
@@ -298,9 +325,9 @@ class Consumer(object):
         except conn.connection.channel_errors as exc:
             # We retries once because of some races that we can
             # recover before informing the deployer
-            # bugs.launchpad.net/oslo.messaging/+bug/1581148
-            # bugs.launchpad.net/oslo.messaging/+bug/1609766
-            # bugs.launchpad.net/neutron/+bug/1318721
+            # https://bugs.launchpad.net/oslo.messaging/+bug/1581148
+            # https://bugs.launchpad.net/oslo.messaging/+bug/1609766
+            # https://bugs.launchpad.net/neutron/+bug/1318721
 
             # 406 error code relates to messages that are doubled ack'd
 
@@ -417,8 +444,6 @@ class ConnectionLock(DummyConnectionLock):
 class Connection(object):
     """Connection object."""
 
-    pools = {}
-
     def __init__(self, conf, url, purpose):
         # NOTE(viktors): Parse config options
         driver_conf = conf.oslo_messaging_rabbit
@@ -443,6 +468,35 @@ class Connection(object):
             driver_conf.kombu_missing_consumer_retry_timeout
         self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
         self.kombu_compression = driver_conf.kombu_compression
+        self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
+
+        if self.heartbeat_in_pthread:
+            # NOTE(hberaud): Experimental: threading module is in use to run
+            # the rabbitmq health check heartbeat. in some situation like
+            # with nova-api, nova need green threads to run the cells
+            # mechanismes in an async mode, so they used eventlet and
+            # greenlet to monkey patch the python stdlib and get green threads.
+            # The issue here is that nova-api run under the apache MPM prefork
+            # module and mod_wsgi. The apache prefork module doesn't support
+            # epoll and recent kernel features, and evenlet is built over epoll
+            # and libevent, so when we run the rabbitmq heartbeat we inherit
+            # from the execution model of the parent process (nova-api), and
+            # in this case we will run the heartbeat through a green thread.
+            # We want to allow users to choose between pthread and
+            # green threads if needed in some specific situations.
+            # This experimental feature allow user to use pthread in an env
+            # that doesn't support eventlet without forcing the parent process
+            # to stop to use eventlet if they need monkey patching for some
+            # specific reasons.
+            # If users want to use pthread we need to make sure that we
+            # will use the *native* threading module for
+            # initialize the heartbeat thread.
+            # Here we override globaly the previously imported
+            # threading module with the native python threading module
+            # if it was already monkey patched by eventlet/greenlet.
+            global threading
+            threading = stdlib_threading
+        self.direct_mandatory_flag = driver_conf.direct_mandatory_flag
 
         if self.ssl:
             self.ssl_version = driver_conf.ssl_version
@@ -535,7 +589,7 @@ class Connection(object):
         # expected waiting the events drain, we start heartbeat_check and
         # retrieve the server heartbeat packet only two times more than
         # the minimum required for the heartbeat works
-        # (heatbeat_timeout/heartbeat_rate/2.0, default kombu
+        # (heartbeat_timeout/heartbeat_rate/2.0, default kombu
         # heartbeat_rate is 2)
         self._heartbeat_wait_timeout = (
             float(self.heartbeat_timeout_threshold) /
@@ -560,7 +614,7 @@ class Connection(object):
 
         # NOTE(sileht): value chosen according the best practice from kombu
         # http://kombu.readthedocs.org/en/latest/reference/kombu.common.html#kombu.common.eventloop
-        # For heatbeat, we can set a bigger timeout, and check we receive the
+        # For heartbeat, we can set a bigger timeout, and check we receive the
         # heartbeat packets regulary
         if self._heartbeat_supported_and_enabled():
             self._poll_timeout = self._heartbeat_wait_timeout
@@ -757,6 +811,10 @@ class Connection(object):
             # NOTE(sileht): we must reraise this without
             # trigger error_callback
             raise
+        except exceptions.MessageUndeliverable:
+            # NOTE(gsantomaggio): we must reraise this without
+            # trigger error_callback
+            raise
         except Exception as exc:
             error_callback and error_callback(exc)
             self._set_current_channel(None)
@@ -770,6 +828,11 @@ class Connection(object):
             LOG.error(msg)
             raise exceptions.MessageDeliveryFailure(msg)
 
+    @staticmethod
+    def on_return(exception, exchange, routing_key, message):
+        raise exceptions.MessageUndeliverable(exception, exchange, routing_key,
+                                              message)
+
     def _set_current_channel(self, new_channel):
         """Change the channel to use.
 
@@ -788,7 +851,8 @@ class Connection(object):
         if new_channel is not None:
             if self.purpose == rpc_common.PURPOSE_LISTEN:
                 self._set_qos(new_channel)
-            self._producer = kombu.messaging.Producer(new_channel)
+            self._producer = kombu.messaging.Producer(new_channel,
+                                                      on_return=self.on_return)
             for consumer in self._consumers:
                 consumer.declare(self)
 
@@ -896,7 +960,7 @@ class Connection(object):
 
     def _heartbeat_start(self):
         if self._heartbeat_supported_and_enabled():
-            self._heartbeat_exit_event = eventletutils.Event()
+            self._heartbeat_exit_event = threading.Event()
             self._heartbeat_thread = threading.Thread(
                 target=self._heartbeat_thread_job, name="Rabbit-heartbeat")
             self._heartbeat_thread.daemon = True
@@ -953,7 +1017,7 @@ class Connection(object):
                                  "occurred, trying to reconnect: %s", exc)
                         self.ensure_connection()
                 except Exception:
-                    LOG.warning("Unexpected error during heartbeart "
+                    LOG.warning("Unexpected error during heartbeat "
                                 "thread processing, retrying...")
                     LOG.debug('Exception', exc_info=True)
 
@@ -1054,13 +1118,12 @@ class Connection(object):
         responses for call/multicall
         """
 
-        # TODO(obondarev): use default exchange since T release
-        consumer = Consumer(exchange_name=topic,
+        consumer = Consumer(exchange_name='',  # using default exchange
                             queue_name=topic,
-                            routing_key=topic,
+                            routing_key='',
                             type='direct',
                             durable=False,
-                            exchange_auto_delete=True,
+                            exchange_auto_delete=False,
                             queue_auto_delete=False,
                             callback=callback,
                             rabbit_ha_queues=self.rabbit_ha_queues,
@@ -1104,7 +1167,7 @@ class Connection(object):
         self.declare_consumer(consumer)
 
     def _ensure_publishing(self, method, exchange, msg, routing_key=None,
-                           timeout=None, retry=None):
+                           timeout=None, retry=None, transport_options=None):
         """Send to a publisher based on the publisher class."""
 
         def _error_callback(exc):
@@ -1113,7 +1176,8 @@ class Connection(object):
                       "'%(topic)s': %(err_str)s", log_info)
             LOG.debug('Exception', exc_info=exc)
 
-        method = functools.partial(method, exchange, msg, routing_key, timeout)
+        method = functools.partial(method, exchange, msg, routing_key,
+                                   timeout, transport_options)
 
         with self._connection_lock:
             self.ensure(method, retry=retry, error_callback=_error_callback)
@@ -1135,7 +1199,8 @@ class Connection(object):
                      'connection_id': self.connection_id})
         return info
 
-    def _publish(self, exchange, msg, routing_key=None, timeout=None):
+    def _publish(self, exchange, msg, routing_key=None, timeout=None,
+                 transport_options=None):
         """Publish a message."""
 
         if not (exchange.passive or exchange.name in self._declared_exchanges):
@@ -1144,21 +1209,25 @@ class Connection(object):
 
         log_info = {'msg': msg,
                     'who': exchange or 'default',
-                    'key': routing_key}
+                    'key': routing_key,
+                    'transport_options': str(transport_options)}
         LOG.trace('Connection._publish: sending message %(msg)s to'
                   ' %(who)s with routing key %(key)s', log_info)
-
         # NOTE(sileht): no need to wait more, caller expects
         # a answer before timeout is reached
         with self._transport_socket_timeout(timeout):
-            self._producer.publish(msg,
-                                   exchange=exchange,
-                                   routing_key=routing_key,
-                                   expiration=timeout,
-                                   compression=self.kombu_compression)
+            self._producer.publish(
+                msg,
+                mandatory=transport_options.at_least_once if
+                transport_options else False,
+                exchange=exchange,
+                routing_key=routing_key,
+                expiration=timeout,
+                compression=self.kombu_compression)
 
     def _publish_and_creates_default_queue(self, exchange, msg,
-                                           routing_key=None, timeout=None):
+                                           routing_key=None, timeout=None,
+                                           transport_options=None):
         """Publisher that declares a default queue
 
         When the exchange is missing instead of silently creates an exchange
@@ -1195,7 +1264,8 @@ class Connection(object):
 
     def _publish_and_raises_on_missing_exchange(self, exchange, msg,
                                                 routing_key=None,
-                                                timeout=None):
+                                                timeout=None,
+                                                transport_options=None):
         """Publisher that raises exception if exchange is missing."""
         if not exchange.passive:
             raise RuntimeError("_publish_and_retry_on_missing_exchange() must "
@@ -1203,7 +1273,7 @@ class Connection(object):
 
         try:
             self._publish(exchange, msg, routing_key=routing_key,
-                          timeout=timeout)
+                          timeout=timeout, transport_options=transport_options)
             return
         except self.connection.channel_errors as exc:
             if exc.code == 404:
@@ -1226,11 +1296,14 @@ class Connection(object):
                                          durable=False,
                                          auto_delete=True,
                                          passive=True)
-
+        options = oslo_messaging.TransportOptions(
+            at_least_once=self.direct_mandatory_flag)
         self._ensure_publishing(self._publish_and_raises_on_missing_exchange,
-                                exchange, msg, routing_key=msg_id)
+                                exchange, msg, routing_key=msg_id,
+                                transport_options=options)
 
-    def topic_send(self, exchange_name, topic, msg, timeout=None, retry=None):
+    def topic_send(self, exchange_name, topic, msg, timeout=None, retry=None,
+                   transport_options=None):
         """Send a 'topic' message."""
         exchange = kombu.entity.Exchange(
             name=exchange_name,
@@ -1240,7 +1313,8 @@ class Connection(object):
 
         self._ensure_publishing(self._publish, exchange, msg,
                                 routing_key=topic, timeout=timeout,
-                                retry=retry)
+                                retry=retry,
+                                transport_options=transport_options)
 
     def fanout_send(self, topic, msg, retry=None):
         """Send a 'fanout' message."""
