@@ -23,6 +23,7 @@ import fixtures
 import kombu
 import kombu.transport.memory
 from oslo_serialization import jsonutils
+from oslo_utils import eventletutils
 import testscenarios
 
 import oslo_messaging
@@ -49,7 +50,7 @@ class TestHeartbeat(test_utils.BaseTestCase):
                                 fake_logger, heartbeat_side_effect=None,
                                 info=None):
 
-        event = threading.Event()
+        event = eventletutils.Event()
 
         def heartbeat_check(rate=2):
             event.set()
@@ -89,6 +90,11 @@ class TestHeartbeat(test_utils.BaseTestCase):
             info='A recoverable connection/channel error occurred, '
             'trying to reconnect: %s')
 
+    def test_run_heartbeat_in_pthread(self):
+        self.config(heartbeat_in_pthread=True,
+                    group="oslo_messaging_rabbit")
+        self._do_test_heartbeat_sent()
+
 
 class TestRabbitQos(test_utils.BaseTestCase):
 
@@ -199,6 +205,7 @@ class TestRabbitPublisher(test_utils.BaseTestCase):
             'msg', expiration=1,
             exchange=exchange_mock,
             compression=self.conf.oslo_messaging_rabbit.kombu_compression,
+            mandatory=False,
             routing_key='routing_key')
 
     @mock.patch('kombu.messaging.Producer.publish')
@@ -212,6 +219,7 @@ class TestRabbitPublisher(test_utils.BaseTestCase):
             conn._publish(exchange_mock, 'msg', routing_key='routing_key')
         fake_publish.assert_called_with(
             'msg', expiration=None,
+            mandatory=False,
             compression=self.conf.oslo_messaging_rabbit.kombu_compression,
             exchange=exchange_mock,
             routing_key='routing_key')
@@ -997,6 +1005,7 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
 
 class ConnectionLockTestCase(test_utils.BaseTestCase):
     def _thread(self, lock, sleep, heartbeat=False):
+
         def thread_task():
             if heartbeat:
                 with lock.for_heartbeat():
